'''Name: ...
Level attempt: HD
'''
import sys

####################### Customer class #############################
class Customer:
    def __init__(self, id_num, name):
        self.id = id_num
        self.name = name
        
    def get_id(self):
        return self.id
    
    def get_name(self):
        return self.name
    
    def get_service_fee(self, cost):
        # In case method is not implemented for the child class,
        # assume 0 service fee
        return 0 
    
    def get_discount(self, cost):
        # In case method is not implemented for the child class,
        # assume 0 discount
        return 0
    
    def display_info(self):
        print(f"Name: {self.get_name()}")
        print(f"Id  : {self.get_id()}")
        print(f"Type: C")
    
    ## for updating txt files 
    def str_helper(self, type):
        return f'id: {self.get_id()}, type: {type}, name: {self.get_name()}'


class BronzeCustomer(Customer):    
    
    service_fee_rate = 0.1
    
    def __init__(self, id_num, name):
        self.service_fee_rate = BronzeCustomer.service_fee_rate
        self.get_discount_rate = 0
        super().__init__(id_num, name)

    ## for updating txt files   
    def __str__(self):
        return super().str_helper('B')

    def get_service_fee_rate(self):
        return BronzeCustomer.service_fee_rate

    def get_service_fee(self, cost):
        return cost*BronzeCustomer.service_fee_rate
    
    def get_discount(self, cost):
        return 0
    
    def display_info(self):   
        print(f"Id: {self.get_id()}")
        print(f"Type: B")
        print(f"Name: {self.get_name()}")
        print(f"Service Fee Rate: {self.get_service_fee_rate()}")
        
   
    def set_service_fee_rate(self, new_rate):
        BronzeCustomer.service_fee_rate = new_rate
     
        
class SilverCustomer(Customer):
    def __init__(self, id_num, name):
        super().__init__(id_num, name)

    ## for updating txt files    
    def __str__(self):
        return super().str_helper('S')

    def get_service_fee(self, cost):
        return 0
    
    def get_discount(self, cost):
        return 0
    
    def display_info(self):        
        print(f"Id: {self.get_id()}")
        print(f"Type: S")
        print(f"Name: {self.get_name()}")
        print(f"Service Fee Rate: 0")
        

        
class GoldCustomer(Customer):
    def __init__(self, id_num, name, discount_rate = 0.08):
        super().__init__(id_num, name)
        self.discount_rate = discount_rate

    ## for updating txt files
    def __str__(self):
        return super().str_helper('G') + f", discount: {self.discount_rate}"

    def get_discount_rate(self):
        return self.discount_rate
        
    def get_service_fee(self, cost):
        return 0
    
    def get_discount(self, cost):
        return cost*self.get_discount_rate()
    
    def display_info(self):      
        print(f"Id: {self.get_id()}")
        print(f"Type: G")
        print(f"Name: {self.get_name()}")
        print(f"Service Fee Rate: 0")
        print(f"Discount Rate: {self.get_discount_rate()}")
        
        
    def set_discount(self, new_discount):
        self.discount_rate = new_discount

######################## Item class #############################             
class Item:
    def __init__(self, id_num, name, price):
        self.id = id_num
        self.name = name
        self.price = price

    def display_info(self):
        print(f"ID: {self.id}")
        print(f"Name: {self.name}")
        print(f"Price: {self.price}")
        
class FoodDish(Item):
    def __init__(self, id_num, name, price):
        super().__init__(id_num, name, price)
    
    ## for updating txt files
    def __str__(self):
        return f"id: {self.id}, type: F, name: {self.name}, price: {self.price}"

    def display_info(self):
        print(f"ID: {self.id}")
        print(f"Type: F")
        print(f"Name: {self.name}")
        print(f"Price: {self.price}")
        
        
class Drink(Item):
    def __init__(self, id_num, name, price):
        super().__init__(id_num, name, price)

    ## for updating txt files
    def __str__(self):
        return f"id: {self.id}, type: D, name: {self.name}, price: {self.price}"

    def display_info(self):
        print(f"ID: {self.id}")
        print(f"Type: D")
        print(f"Name: {self.name}")
        print(f"Price: {self.price}")

class Banquet(Item):
    def __init__(self, id_num, name, components, price):
        self.components = components
        super().__init__(id_num, name, price)

    ## for updating txt files
    def __str__(self):
        line = " ".join([x.name for x in self.components])
        return f"id: {self.id}, type: B, name: {self.name}, components: {line}"

    def display_info(self):
        print(f"ID: {self.id}")
        print(f"Type: B")
        print(f"Name: {self.name}")
        print(f"Price: {self.price}") 
        components = [i.name for i in self.components]
        print("Components: ", end = "")
        for i in components:
            print(i, end = " ")
        print()
        
##################################### Order class########################################    
class Order:
    def __init__(self, cust, item, quantity):
        self.cust = cust
        self.item = [item] if isinstance(item, Item) else item
        self.quantity = quantity if isinstance(quantity, list) else [quantity]
    
    ## for updating txt files
    def __str__(self):
        zipped = zip(self.item, self.quantity)
        combined = [f"{i.name} x {q}" for i, q in zipped]
        return ", ".join(combined)

    def compute_cost(self):
        orig_cost = sum([x.price*y for x,y in zip(self.item, self.quantity)])
        service_fee = self.cust.get_service_fee(orig_cost)
        discount = self.cust.get_discount(orig_cost)
        return round(orig_cost, 2), round(service_fee, 2), round(discount ,2)
    
    def compute_final_cost(self):
        a,b,c = self.compute_cost()
        return a+b-c

    def display_info(self):
        print(f"Name : {self.cust.get_name()}")
        print(f"Items: {' '.join([x.name for x in self.item])}")
        print(f"Quantity: {' '.join(map(str, self.quantity))}")
    

############## Records #################################
class Records:
    def __init__(self, cust_list = None, item_list = None):
        # Will store the object itself rather than the IDs or the names
        self.customer_list = [] if cust_list is None else cust_list
        self.item_list = [] if item_list is None else item_list
        self.order_list = []
        if cust_list is not None:
            self.max = max([cust.id for cust in cust_list], default = 0)
        else: self.max = 0
        
    def read_customers(self, file_name):
        try:
            with open(file_name, "r") as f:
                lines = [x.strip().split(',') for x in f.readlines()]
        except FileNotFoundError:
            return False
        
        def read_customer_as_line(customer):
            id_num = int(customer[0].split(":")[1])
            cust_type = customer[1].split(":")[1].strip()
            cust_name = customer[2].split(":")[1].strip()
            if cust_type == "G":
                discount = float(customer[3].split(": ")[1])
                return GoldCustomer(id_num, cust_name, discount)
            if cust_type == "S":
                return SilverCustomer(id_num, cust_name)
            return BronzeCustomer(id_num, cust_name)
        
        new_customers = map(read_customer_as_line, lines)
        
        self.customer_list.extend(new_customers)
        self.max = max([cust.id for cust in self.customer_list], default = 0)
        return True #Indicating File reading was successful
        
    def read_items(self, file_name):
        try:
            with open(file_name, "r") as f:
                lines = [x.strip() for x in f.readlines()]
        except FileNotFoundError:
            return False
        non_banuqet = [x.split(',') for x in lines if 'type: B' not in x]
        banuqet = [x.split(',') for x in lines if 'type: B' in x]
        
        def read_item_as_line(item):
            id_num = int(item[0].split(":")[1])
            item_type = item[1].split(":")[1].strip()
            item_name = item[2].split(":")[1].strip()

            price = float(item[3].split(":")[1])

            if item_type == "F":
                return FoodDish(id_num, item_name, price)
            elif item_type == "D":
                return Drink(id_num, item_name, price)

        new_items = map(read_item_as_line, non_banuqet)
        self.item_list.extend(new_items)

        
        def validBanquet(components):
            #check if any food item is in Banquet
            food_exists = any(map(FoodDish.__instancecheck__, components))
            #check if any drink item is in Banquet
            drink_exists = any(map(Drink.__instancecheck__, components))
            return food_exists and drink_exists
        
        def read_banquet_as_line(item):
            id_num = int(item[0].split(":")[1])
            item_type = item[1].split(":")[1].strip()
            item_name = item[2].split(":")[1].strip()

            components = item[3].split(":")[1].strip()
            components = [self.find_item(i) for i in components.split(" ")]
            components = [x for x in components if x is not None]
            if not validBanquet(components):
                print(f"something wrong with Banquet {item_name}")
                return None
            price = sum([i.price for i in components])

            return Banquet(id_num, item_name, components, price)

        banuqet_items = map(read_banquet_as_line, banuqet)
        banuqet_items = [x for x in banuqet_items if x is not None]
        self.item_list.extend(banuqet_items)  

        return True #Indicating File reading was successful

    def find_customer(self, value):
        try:
            value = int(value)
        except ValueError:
            pass
        for cust in self.customer_list:
            if cust.get_id() == value or cust.get_name() == value:
                return cust
        return None

    def find_item(self, value):
        try:
            value = int(value)
        except ValueError:
            pass
        for item in self.item_list:
            if item.id == value or item.name == value:
                return item
        return None

    def display_customers(self):
        for cust in self.customer_list:
            cust.display_info()
            print()

    def display_items(self):
        for item in self.item_list:
            item.display_info()
            print()

    def display_orders(self):
        for item in self.order_list:
            item.display_info()
            print()

    def update_customer_file(self, filepath):
        to_write = "\n".join([str(x) for x in self.customer_list])
        with open(filepath, 'w') as file:
            file.write(to_write)

    def update_item_file(self, filepath):
        non_banquet = [x for x in self.item_list if not isinstance(x, Banquet)]
        banquet = [x for x in self.item_list if not isinstance(x, Banquet)]

        with open(filepath, 'w') as file:
            file.write("\n".join(non_banquet))
            file.write("\n")
            file.write("\n".join(banquet))

################################Operations class##############################
class Operations:
    def __init__(self, customer_file, item_file):
        self.records = Records()
        if not self.records.read_customers(customer_file) or not self.records.read_items(item_file):
            # print("Existing Records not found. Exitting...")
            sys.exit("Existing Records not found. Exitting...")
        self.main_loop()

    def main_loop(self):
        print("Welcome to RMIT restaurant ordering system!")
        print()
        print("############################################################")
        print("1. Order a meal")
        print("2. Display existing customers information")
        print("3. Display existing items information")
        print("4. Order meals via a file")
        print("5. Adjust Bronze Service Fee")
        print("6. Adjust discount of a Gold Member")
        print("7. Display all Orders")
        print("8. Display the most frequently customer")
        print("9. Display a customer's order history")
        print("0. Quit")
        print("############################################################")
        print()
        
        choice = None
        while choice is None:
            try:
                choice = int(input("Choose one option: "))
            except ValueError:
                print("Enter a valid choice")
                choice = None
        
        match choice:
            case 0:
                sys.exit("Exitting...")
            case 1:
                self.order()
            case 2: 
                self.records.display_customers()
            case 3: 
                self.records.display_items()
            case 4:
                self.order_meals_via_file()
            case 5:
                new_fee = None
                while new_fee is None:
                    try:
                        new_fee = float(input("Enter new default Service Fee: "))
                        if new_fee <= 0: raise ValueError
                    except ValueError:
                        print("Enter a number greater than 0")
                        new_fee = None
                BronzeCustomer(1,1).set_service_fee_rate(new_fee)
            case 6:
                cust = None
                while cust is None:
                    cust = self.records.find_customer(input("Enter Customer name or ID: "))
                    if not isinstance(cust, GoldCustomer):
                        print("Customer not found or not a Gold Customer")
                        cust = None
                
                new_discount = None
                while new_discount is None:
                    try:
                        new_discount = float(input("Enter new discount: "))
                        if new_discount <= 0: raise ValueError
                    except ValueError:
                        print("Enter a number greater than 0")
                        new_discount = None
                
                cust.set_discount(new_discount)

            case 7:
                self.records.display_orders()

            case 8: #display most freq customer
                orders = self.records.order_list
                cust = [x.cust.name for x in orders]
                Max, freq_cust  = 0, None 
                d = {}
                for i in cust:
                    if i in d:
                        d[i] += 1                      
                    else:
                        d[i] = 1

                    if Max < d[i]: 
                        Max = d[i]
                        freq_cust = i
                print(f"Most Frequent customer: {freq_cust} with {Max} orders" ) 
                print()

            case 9: #display customer's order history
                self.display_customer_history()

            case _:
                print("Please Enter valid choice")
        
        self.main_loop()

    def display_customer_history(self):
        cust = self.records.find_customer(input("Please enter customer name or ID: "))
        if not isinstance(cust, Customer):
            print("Customer not found. Returning to main menu.")
            return

        cust_orders = [order for order in self.records.order_list if order.cust is cust]
        print(f"{' ': <30}{'Order': <40}{'Price':^20}")
        for i, order in enumerate(cust_orders):
            serial = f"Order {i+1}"
            print(f"{serial :<30}{str(order): <40}{order.compute_final_cost(): ^20}")
        print()


    def order_meals_via_file(self):
        filepath = input("Enter File name: ")

        try:
            with open(filepath, "r") as f:
                lines = [x.strip().split(',') for x in f.readlines()]
        except FileNotFoundError:
            print("File not found. Going back to main menu.")
            self.main_loop()

        def read_order_from_line(order):
            name = order[0].split(':')[1].strip()
            customer = self.records.find_customer(name)

            items = order[1].split(':')[1].strip().split(' ') 
            item_names = items[::2]
            item_quan = list(map(int, items[1::2]))
            items = [self.records.find_item(x) for x in item_names]

            return Order(customer, items, item_quan)

        orders = map(read_order_from_line, lines)
        self.records.order_list.extend(orders)
        print("Orders have been placed sucessfully!")
        print()
                
    def order(self):
        name = input("Enter the name of the customer [e.g. Huong]: ")
        
        items = []
        while True:
            item = input("Enter the item name or ID: ")
            item = self.records.find_item(item)
            while not item:
                print("Invalid item")
                item = input("Enter the item [enter a valid item only, e.g. hamburger, coke, pizza]: ")
                item = self.records.find_item(item)
                
            quantity = None
            while not quantity:
                quantity = input("Enter the item quantity [enter a positive integer only, e.g. 1, 2, 3]: ")
                try:
                    quantity = int(quantity)
                    if quantity <= 0: raise ValueError
                except ValueError:
                    print("Enter an Postive Integer")
                    quantity = None

            items.append((item, int(quantity)))

            choice = input("Does the customer want more items?")
            while choice != 'y' and choice != 'n':
                choice = input("Please enter 'y' or 'n': ")
            
            if choice == 'n': break

        cust = self.records.find_customer(name)
        reg_fee = 0
        # if customer is new
        if not cust:
            new = input("This is a new customer. Does the customer want to join the rewards program [enter y or n]? ")
            while new != 'y' and new != 'n':
                new = input("Please enter 'y' or 'n': ")

            if new == "y":
                reward = input("What kind of rewards the customer wants? ")
                while reward != "S" and reward != "G":
                    reward = input("Enter valid reward category ('S' or 'G'): ")
            else:
                reward = "B"
            
            #adding new customer info to the records
            self.records.max += 1
            match reward:
                case "B":
                    cust = BronzeCustomer(self.records.max, name)
                case "S":
                    cust = SilverCustomer(self.records.max, name)
                    reg_fee = 100
                case "G":
                    cust = GoldCustomer(self.records.max, name)
                    reg_fee = 300
                    
            self.records.customer_list.append(cust)        
        
        order = Order(cust, [x[0] for x in items], [x[1] for x in items])
        self.records.order_list.append(order)
        orig_cost, service_fee, discount = order.compute_cost()
        
        ## printing bill
        print("*"*50)
        print(f"Receipt of Customer {name}")
        print("*"*50)
        for i,q in items:
            l = i.name + ":"
            r = f"{i.price} (AUD) x {q}"
            print(f"{l :<30}{r :>20}")
        l, r = "Service fee:", f"{service_fee} (AUD)"        
        print(f"{l :<30}{r :>20}")
        if discount > 0:
            l, r = "Discount: ", f"{discount} (AUD)"
            print(f"{l :<30}{r :>20}")
        if reg_fee > 0:
            l, r = "Registration fee: ", f"{reg_fee} (AUD)"
            print(f"{l :<30}{r :>20}") 
        l, r = "Total cost: ", f"{ round(orig_cost + reg_fee + service_fee -discount, 2)} (AUD)"
        print(f"{l :<30}{r :>20}")
        print("*"*50)
        print()

#############################################################################################
if __name__ == "__main__":
    import sys
    n = len(sys.argv)
    if n != 3 and n != 1:
        print(f"Incorrect number of arguments entered. Please enter 2 arguments. For example 'python script.py customer.txt item.txt'")
        sys.exit("Exitting...")
    if n == 3:
        customer_file = sys.argv[1]
        item_file = sys.argv[2]
    else:
        customer_file = 'customers.txt'
        item_file = 'items.txt'
    Operations(customer_file, item_file)

    
